module HF

  WrongSize = Class.new(StandardError)
  WrongType = Class.new(StandardError)
  WrongArgumentType = Class.new(StandardError)

  module ::Enumerable

    SINGLE_SIZE = 1

    # @return [Boolean] true when enum is a 1-tuple or single
    def single?
      return true if self.count == SINGLE_SIZE
      return false
    end
  
    # Returns the only element in a 1-tuple/single.
    # A more idiomatic and safer accessor than first when
    # the Enumerator is expected to have only one element.
    # @raise [WrongSize] when the enumerator is not a 1-tuple
    # @return [Object] the element in the 1-tuple enumerator
    def only
      return self.first if self.single?
      raise WrongSize, wrong_size_message(self.count, SINGLE_SIZE)
    end

    # @note Uses HF::Range
    # Returns the range of the first one-member multiset collecting
    # the specified value at the given multiplicity. This is a collection
    # such as [3, 3, 3 ,3], where there is multiple instances of one member.
    # Formally it is a multiset with only one member.
    # The default multiplicity returns the range of an mset of arbirtrary size.
    # The default value is not specified it returns the range of the first mset with distinct values.
    # @param value [Object] the element in the mset
    # @param multiplicity [Integer] the number of times value appears in mset
    # @return [Range] the first range where the subset occurs
    def find_mset(value, multiplicity=-1)
      is_multiplicity_negative = multiplicity < 0
      enums_chunked_on_value = self.chunk{|i| i == value}
      sub_enums = enums_chunked_on_value.collect{|r| r.last}
      sub_enums_is_mset =
        if is_multiplicity_negative
          enums_chunked_on_value.collect{|r| r.first && not(r.single?) }
        else
          enums_chunked_on_value.collect{|r| r.first && (r.last.size >= multiplicity) }
        end
      first_mset_chunked_index = sub_enums_is_mset.find_index(true)
      return Range.new(self.count, self.count, true) if first_mset_chunked_index.nil?
      first_mset_offset = sub_enums.take(first_mset_chunked_index).flatten.size
      first_mset_size = is_multiplicity_negative ? sub_enums[first_mset_chunked_index].size : multiplicity
      Range.new(0, first_mset_size.pred).bump(first_mset_offset)
    end

    private

    # @param expected_size [Integer] 
    # @param actual_size [Integer]
    # @return [String] the message for the WrongSize exception
    def wrong_size_message(expected_size, actual_size)
      elm_word = actual_size == 1 ? "element" : "elements"
      "Enumerator is not a #{expected_size}-tuple. It has #{actual_size} #{elm_word} instead of #{expected_size}"
    end
  end

  class ::Hash

    DEFAULT_EMBED_AS = nil

    # Add field to hash as is or embedded under a key given by embed_as. Addition
    # only occurs when field contains data.
    # @param field [Hash] The fields to add to the hash
    # @param embed_as [#to_s] The key under which the field should be embedded
    #   in the hash
    # @return [Hash] replace existing hash with modified hash
    def add_optional!(field, embed_as: DEFAULT_EMBED_AS)
      return self.merge!(field) if embed_as.eql?(DEFAULT_EMBED_AS)
      return self if field.empty?
      self.merge!({embed_as => field})
    end

    # Add field to hash as is or embedded under a key given by embed_as. Addition
    # only occurs when field contains data.
    # @param field [Hash] The fields to add to the hash
    # @param embed_as [#to_s] The key under which the field should be embedded
    #   in the hash
    # @return [Hash] creates a new has with modifications
    def add_optional(field, embed_as: DEFAULT_EMBED_AS)
      return self.merge(field) if embed_as.eql?(DEFAULT_EMBED_AS)
      return self.clone if field.empty?
      self.merge({embed_as => field})
    end

    # @param keys [HashKeyArray] 
    # @return [Hash] a Hash containing only the keys given.
    def slice(*keys)
      self.select{|k,_| keys.include?(k)}
    end

    # @param (see #slice)
    # @return [Hash] a Hash containing none of the keys given
    def drop(*keys)
      self.reject{|k,_| keys.include?(k)}
    end
  end

  class ::Integer

    # Return an IntegerRange that captures all the values after the integer
    # @return [IntegerRange] a new Range
    def after
      (1 + self)..-1
    end

    # Return an IntegerRange that captures all the positive values before the integer
    # @return [IntegerRange] a new Range
    def before
      0..(self - 1)
    end
  end

  class ::Range

    # Bump the beginning and end of an integer range by n
    # @param n [Integer] the amount to raise both begin and end by
    # @return [IntegerRange] a new Range
    def bump(n)
      raise(WrongType, wrong_type_message(Integer)) unless self.begin.is_a?(Integer)
      raise(WrongArgumentType, wrong_argument_type_message(Integer, n.class)) unless n.is_a?(Integer)
      Range.new(self.begin + n, self.end + n, self.exclude_end?)
    end

    # Return an IntegerRange that captures all the values after the end of the interval
    # @return [IntegerRange] a new Range
    def after
      raise(WrongType, wrong_type_message(Integer)) unless self.begin.is_a?(Integer)
      (1 + self.end)..-1
    end

    # Return an IntegerRange that captures all the values before the beginning of the interval
    # @return [IntegerRange] a new Range
    def before
      raise(WrongType, wrong_type_message(Integer)) unless self.begin.is_a?(Integer)
      0..(self.begin - 1)
    end

    private

    # @param expected_type [ClassName] 
    # @return [String] the message for the WrongType exception
    def wrong_type_message(expected_type)
      "Range must have #{expected_type} values and not #{self.begin.class}"
    end

    # @param expected_type [ClassName]
    # @param actual_type [ClassName] 
    # @return [String] the message for the WrongArgumentType exception
    def wrong_argument_type_message(expected_type, actual_type)
      "Argument must be an #{expected_type} and not a #{actual_type}"
    end
  end
end
